(* :Title: Manager *)

(* :Authors: V.N.Larin, V.V.Ezhela *)

(* :Summary:
This package is management system for the special database of the
physical constants "DBFPC" including three lists:  'basicTable', 
'standardNames', 'ccTable' (correlation coefficient table). It
allows to execute different manipulations with the constant data
and to add new data. It includes also special modules for 
uncertainty calculation for various physical constant 
expressions with usage of the correlation coefficients - 
"ErrorPropagator" and for the covariances, variances and
correlation coefficients calculation - "Covariator".
*)

(* :Package Version: 2.1 *)

(* :Copyright: Copyright 2000-2005, State Research Center Of Russia
                                    "Institute For High Energy Physics". *)

(* :Context: StandardPhysicalConstants`Manager` *)

(* :History:
      Version 1.0: original version by V.N.Larin, V.V.Ezhela (SRCR IHEP),
                   18.03.2000.
      Version 1.1: by V.N.Larin, V.V.Ezhela (SRCR IHEP), 04.07.2002.
      Version 1.2: by V.N.Larin, V.V.Ezhela (SRCR IHEP), 11.03.2003.
      Version 1.3: by V.N.Larin, V.V.Ezhela (SRCR IHEP), 11.06.2003.
      Version 2.0: by V.N.Larin, V.V.Ezhela (SRCR IHEP), 24.11.2003.
      Version 2.1: by V.N.Larin, V.V.Ezhela (SRCR IHEP), 26.01.2004.

                   Last revision:  26.01.04.
*)

(* :Keywords: *)

(* :Source: *)

(* :Warning: Redefines variables: Mole,mbarn,fMeter,cMeter,kMeter,
             Meter,kParsec,Gram,kGram,Joule,Weber,Kelvin,Ampere,
             Coulomb,eV,MeV,GeV,Volt,Tesla,Watt,Hertz,kHertz,
             GYear *)

(* :Mathematica Version: 4.0 *)

(* :Limitation: Unknown. *)

(* :Discussion: *)

(* ================================================================ *)
Print[StyleForm["   Please, don't redefine symbols:",
                             FontSize->11,
                             FontSlant->"Italic",
                             FontWeight->"Bold",
                             FontFamily->"Helvetica",
                             FontColor->RGBColor[0.999,0.,0.]
]];
Print[StyleForm["   dataBase, basicTable, standardNames, ccTable",
                             FontSize->11,
                             FontSlant->"Italic",
                             FontWeight->"Bold",
                             FontFamily->"Helvetica",
                             FontColor->RGBColor[0,0.5,0.999]]];

BeginPackage["StandardPhysicalConstants`Manager`"]

ErrorPropagator::usage =
"ErrorPropagator[expression,database,{control_keys}] calculates mean value,
uncertainty and relative uncertainty for constant expression. It carries out
test on positive definite correlation submatrix, allows to control
by modification of correlation submatrix and output."

Covariator::usage =
"Covariator[expr_List,datbas,cntrl_List] gives the covariances, variances
and correlations of the estimated constants in terms of the covariances
of the 'adjusted' constants."

MatrixCharacteristics::usage =
"MatrixCharacteristics[matr_List, heading_String] calculates some characteristics
numerical matrix 'matr': condition number, determinant, elements rounding
criterion (by VVE), eigenvalues."

UM::usage = 
"UM[expr] realizes canonical statement of the units of identical physical
quantities;  UM[expr,goal] expresses all SI units in terms of the main
SI units (under study)."

SetPower::usage =
"SetPower[number,power] present number in standard form with setting power
(under study)."

Access::usage =
"Access[constant_table,constant_name] output the constant data:
{mean_value,uncertainty,relative_uncertainty} as standard list."

ValueSyntax::usage =
"ValueSyntax[one_string_of_constant_table] execute syntax analysis of
the constant value in constant table format."

MakeCM::usage =
"MakeCM[db]; MakeCM[db,\"print\"] make correlation matrix corresponding
to the correlation coefficient table from 'db'."

PairCC::usage =
"PairCC[nc1,nc2,db] output the correlation coefficient for pair of
constants: 'nc1' and 'nc2' from 'db'."

GroupCC::usage =
"GroupCC[nc1,nc2,db] output the correlation coefficients for
constant group from 'db' ('nc1' - first constant, 'nc2' - list
of other constants)."

InsertFPC::usage =
"InsertFPC[{data_const_list}, database, ccdef] inserts new constant into
the data base 'database'; ccdef - initial value of correlation coefficients
(default value: ccdef=\" \" for InsertFPC[{data_const_list}, database])."

StringAO::usage =
"StringAO[in_string,string_list] gives position number for in_string
in string_list (in alphabetical order)."

DeleteFPC::usage =
"DeleteFPC[nc, database] deletes one constant (nc - name or code) or list
of the constants (nc - list of the names/codes) from the data base."

Copy::usage =
"Copy[clis, dbs, dbr] makes the copies of the constants (clis) from the 
source data base (dbs) to the receiver data base (dbr)."

Shift::usage =
"Shift[name_or_code_const,new_code,database] transfers constant
attributes at new positions in the all tables."

NewListFPC::usage =
"NewListFPC[newlis,db] create the new (sub-) data base corresponding
to list of the constants 'newlis' (from 'db')."

PrintDB::usage =
"PrintDB[db] print current data base db={ct,cnt,cct} or separate table
db[[i]] to the channel $Output (terminal screen by default)."

PrintBT::usage =
"PrintBT[ct] print constant table to the channel $Output
(terminal screen by default)."

PrintSN::usage =
"PrintSN[cnt] print constant name table to the channel
$Output (terminal screen by default)."

PrintCC::usage =
"PrintCC[cct] print correlation coefficient table to the channel
$Output (terminal screen by default)."

OutputDB::usage =
"OutputDB[database] outputs a created data base to the file with name
Dbfpc_n.m (in the current directory)."

Unprotect[ErrorPropagator,Covariator,MatrixCharacteristics,UM,SetPower,
          Access,ValueSyntax,MakeCM,PairCC,GroupCC, InsertFPC,InsertDB,
          StringAO,DeleteFPC,DeleteDB,Copy,Shift,NewListFPC,
          PrintDB,PrintBT,PrintSN,PrintCC,OutputDB, cC,
          mbarn,fMeter,kMeter,cMeter,Meter,Gram,kGram,Joule,eV,MeV,GeV,
          Watt,Ampere,Coulomb,Volt,Tesla,Weber,Hertz,kHertz,
          Mole,Kelvin,kParsec,GYear];

Begin["`Private`"]

 ErrorPropagator[expr_, database:{___List}, control___List] :=
  
      Module[ {form=expr, bt=First[database], cct=Last[database],
               var, svar, eps, dim, vr, svr, es, dm, cc, tc, k, ru, val,
               facs, vform, unc,aunc,coef,result,
               ctr=control,ckey=Null,okey=Null,ncp, unit,du,erun,
               ermxx,cd,k1,k2, flag=0,nuls=0,cco,ermo,valr,mineig},

If[ctr === Null, Goto[n0], ncp = First[Dimensions[ctr]] ];
ckey = ctr[[1]];
If[ncp>1, okey = ctr[[2]];
          If[ncp>2, Print[TableForm[
            {"ErrorPropagator: Only 2 elemets may be in last parameter"}]]]];
Label[n0];
               var = Variables[form];
               dim = First[Dimensions[var]];
               If[ dim==0, result={form,Null,Null}; Goto[end] ];
               If[ dim >1, 
               var = Union[Select[Level[form,Depth[form]], Depth[#]==1 &&
                                                         !NumberQ[#] &]] ];
               var = Select[var,MemberQ[bt[[All, 2]], SymbolName[#]] &];
               dim = First[Dimensions[var]];
               If[ dim==0, result={form,Null,Null}; Goto[end] ];
               svar = SymbolName /@ var;
               eps = Map[Access[bt,#][[2]] &, svar];
               vr = var; svr = svar; es = eps; dm = dim;
               i = 1;
Label[nxt]; 
               If[i>dm, Goto[endif],
                  If[es[[i]]=="Exact", ( es = Delete[es,i];
                     svr = Delete[svr,i]; vr = Delete[vr,i];
                     dm -= 1; Goto[nxt] ),
                                  i += 1; Goto[nxt],
                                  i += 1; Goto[nxt]] ];
Label[endif];
              If[ dm==0, erm={0}; unc=0.; Goto[exactonly] ];
              cc = 1.*IdentityMatrix[dm];
              If[ dm==1, okey=Null; Goto[n1] ];
              If [ckey==0, Goto[n1]];
              Do[ tc = Last[PairCC[svr[[i]], svr[[j]], database]];
                  If[ tc=="Null"||tc==" ",
                      If[ ckey=="Null", m=i; n=j;  If[i>j, m=j; n=i];
                                        nuls=1; cc[[i,j]]=1.*cC[m,n],
                                        Continue[], Continue[] ],
                      cc[[i,j]]=ToExpression[tc] ],
                                                 {i,1,dm}, {j,1,dm}];
              If[ Head[ckey] =!= List, Goto[n1] ];
                  cd = First[Dimensions[ckey]];
                  Do[ k1=ckey[[i,1]];  If[ NumberQ[k1] == True,
                      If[ k1<=dm, Goto[nxt],  Goto[er1] ],
                          Do[ If[ vr[[j]]==k1, k1=j;  Goto[nxt] ],
                                                       {j,1,dm} ] ];
Label[er1];  Print["No such 1-st constant (",k1,") in pair ",i];
                                                     Continue[];
Label[nxt];
                     k2=ckey[[i,2]];
                     If[ NumberQ[k2] == True,
                     If[ k2<=dm, Goto[subs], Goto[er2] ],
                         Do[ If[ vr[[j]]===k2, k2=j; Goto[subs] ],
                                                       {j,1,dm} ] ];
Label[er2];  Print["No such 2-d constant (",k2,") in pair ",i];
                                                     Continue[];
Label[subs]; cc[[k1,k2]] = ckey[[i,3]];
             cc[[k2,k1]] = ckey[[i,3]], {i,cd} ];
Label[n1];
             erm = Table[D[form,vr[[i]]] D[form,vr[[j]]]*
                         es[[i]]*es[[j]]*cc[[i,j]], {i,dm},{j,dm}];
Label[exactonly];
               val  = Map[First[Access[bt,#]] &, svar];
               facs = Table[var[[i]] -> val[[i]], {i, dim}];
               erm  = erm /. facs;
               er   = Flatten[erm];
               If[ dm>0, unc = PowerExpand[Sqrt[Plus @@ er]] ];
               vform = form /. facs;
               ru = If[ vform==0., "", If[ unc!=0, unc/vform, 0, unc/vform ],
                                       If[ unc!=0, unc/vform, 0, unc/vform ] ];
unit=Variables[erm];  unit=DeleteCases[unit, cC[__]];
du=First[Dimensions[unit]]; erun=Table[unit[[i]]->1, {i,du}];
ermxx=erm /.erun;

               coef = If[ NumberQ[unc], unc, Select[unc, NumberQ] ];
               aunc = If[ Head[coef]===Complex, unc, unc*Sign[coef] ];
               ru   = If[ Head[ru]===Complex, ru, Abs[ru] ];
               valres = {vform, aunc, ru};
               If[ dm == 0,
                   Print[" EP:  No correlation matrix (exact constants only)"];
                                                  result = valres; Goto[end] ];
               If[ flag==0, cco=cc; ermo=ermxx,
                            valres={valr, valres}; Goto[ret] ];
               If[ nuls!=0, Goto[ret] ];
               mineig=Min[Eigenvalues[cco]];
               If[ mineig>0, Goto[ret] ];
  Print["  WARNING:  correlation matrix is non-positive semidefinite."];
  Print["            Minimum eigenvalue of this matrix is  ",mineig];
                flag = 1;
                ckey = 0;
                If[ okey===Null, okey="CM",
                                   If[ okey=="EM", okey="All" ] ];
                valr = valres;
                Goto[endif];
Label[ret];
                valres = PowerExpand[UM[valres]];
 
If[okey === Null,       result = valres,
   If[okey == "All",    result = {cco, ermo, valres},
      If[okey == "CM",  result = {cco, valres},
      If[okey == "EM",  result = {ermo,valres},
                        result = valres;
      Print[TableForm[{"EP: OUTPUT-keys: \"All\", \"CM\", \"EM\""}]] ]]]];

Label[end];
      Return[result] ]


 Covariator[expr:{___List},datbas:{___List},cntrl___List]:=
 
      Module[ {sexpr,form1,svr1,sig1,form2,df, bt=First[datbas],ctr=cntrl,
               ckey=Null,okey=Null,ncp,varx,var={},svar,eps,dim,vr,svr,es,
               dm,cc, uniq, formprod, ccout,varian,covar,varcov, mixcm={},
               mix,mixo, val,facs,fact,head=Null, comb,tmpdb, cd,k1,k2,mch,
               pos,flag=0,nuls=0,cco,vrcv,result={}},

      df = First[Dimensions[expr]];
      If[ df==1, result=ErrorPropagator[expr[[1,2]],datbas];
                 head="ErrorPropagator:"; Goto[end] ];
      If[ ctr === Null, Goto[n0], ncp=Length[ctr] ];
      ckey=ctr[[1]];
      If[ncp>1, okey=ctr[[2]];
         If[ncp>2, Print[TableForm[
                   {"Covariator: Two control parameters may be only"}]]]];
Label[n0];
      sexpr = Sort[expr];
      form1 = sexpr[[All,1]];
      form2 = sexpr[[All,2]];
      svr1 = SymbolName /@ form1;
           i = 1;
Label[n2]; If[ i>df, Goto[n3] ];
           If[ FreeQ[bt[[All,2]],svr1[[i]]],
               (Print["Covariator: No output constant \"",form1[[i]],
                      "\" in data base"]; form1=Drop[form1,{i}];
                form2=Drop[form2,{i}];svr1=Drop[svr1,{i}];df-=1;Goto[n2]),
                i+=1; Goto[n2] ];
Label[n3]; 
If[ df==0, Goto[end], If[ df==1, result=ErrorPropagator[form2[[1]],datbas];
                                 head="ErrorPropagator:"; Goto[end] ] ];
      sig1 = Map[Access[bt,#][[2]] &, svr1];
      Do[ varx = Variables[form2[[i]]]; dim = First[Dimensions[varx]];
          If[ dim==0, Continue[] ];
          If[ dim >1, 
          varx = Union[Select[Level[form2[[i]],Depth[form2[[i]]]],
                                    Depth[#]==1 && !NumberQ[#] &]] ];
          varx = Select[varx,MemberQ[bt[[All, 2]], SymbolName[#]] &];
          var=Union[var,varx], {i,df}];
          dim = First[Dimensions[var]];
          If[ dim==0, result=form2;
              Print["No input constants in data base"]; Goto[end] ];
Print[" Input constants:  ",var];
Print[" Output constants:  ",form1];
          svar = SymbolName /@ var;
          eps = Map[Access[bt,#][[2]] &, svar];
          vr = var; svr = svar; es = eps; dm = dim;

               i = 1;
Label[nxt]; 
          If[ i>dm, Goto[endif],
              If[ es[[i]]=="Exact", ( es = Delete[es,i];
                  svr = Delete[svr,i]; vr = Delete[vr,i];
                  dm -= 1; Goto[nxt] ), i += 1; Goto[nxt],
                                        i += 1; Goto[nxt]] ];
Label[endif];
          If[ dm==0, result=form2;
              Print["Input constants are exact"]; Goto[end] ];
Label[flg];
          If[ ckey==0, cc = 1.*IdentityMatrix[dm]; Goto[n1] ];
          tmpdb = NewListFPC[svr, datbas];
          tmpdb[[3]]= tmpdb[[3]];
          cc = MakeCM[tmpdb,"print"];
          Clear[tmpdb];

              Do[ If[ cc[[i,j]]==Null||cc[[i,j]]==" ",
                      If[ ckey=="Null", m=i; n=j;  If[i>j, m=j; n=i];
                                        nuls=1; cc[[i,j]]=1.*cC[m,n],
                                        cc[[i,j]]=0.; Continue[],
                                        cc[[i,j]]=0.; Continue[] ] ],
                                                      {i,1,dm}, {j,1,dm}];
            If[ Head[ckey] =!= List, Goto[n1] ];
            cd = Length[ckey];
            Do[ k1=ckey[[i,1]]; If[ NumberQ[k1],
                                    If[ k1<=dm, Goto[nxt] ],
                                    pos=Position[vr,k1];
                                    If[ pos=!={}, k1=First[Flatten[pos]];
                                                  Goto[nxt] ] ];
Print["Constant ",k1," (in pair ",i,
      ") is absent in list of input constants"];
                                                  Continue[];
Label[nxt];
                k2=ckey[[i,2]]; If[ NumberQ[k2],
                                    If[ k2<=dm, Goto[subs] ],
                                    pos=Position[vr,k2];
                                    If[ pos=!={}, k2=First[Flatten[pos]];
                                                  Goto[subs] ] ];
Print["Constant ",k2," (in pair ",i,
      ") is absent in list of input constants"];
                                                  Continue[];
Label[subs]; If[ !NumberQ[ckey[[i,3]]], nuls=1 ];
             cc[[k1,k2]] = ckey[[i,3]];
             cc[[k2,k1]] = ckey[[i,3]], {i,cd} ];
Label[n1];
If[ flag==0 && nuls==0 && dm!=1, 
    mch = MatrixCharacteristics[cc,"input correlation matrix"] ];

               svar = Union[svar,svr1];
               val  = Map[First[Access[bt,#]] &, svar];
               var  = Union[var,form1];
 comb = UM[{es,val},1];
 es = First[comb];
 val = Last[comb];
 Drop[comb,2];
               dim = First[Dimensions[var]];
               facs = Table[var[[i]] -> val[[i]], {i, dim}];

   varian = DiagonalMatrix[Table[Sum[D[form2[[k]],vr[[i]]] D[form2[[k]],vr[[j]]]*
                          es[[i]]*es[[j]]*cc[[i,j]], {i,dm},{j,dm}]/form1[[k]]^2,
                                                                       {k,df}] ];
  uniq = Unique[];
  covar = Table[If[ k==l, 0, formprod=form1[[k]]*form1[[l]];
                             formprod=formprod*Sign[Select[formprod*uniq/.facs,
                             NumberQ]];
                             Sum[D[form2[[k]],vr[[i]]] D[form2[[l]],vr[[j]]]*
                             es[[i]]*es[[j]]*cc[[i,j]],{i,dm},{j,dm}]/formprod],
                             {k,df},{l,df}];
  varcov = varian + covar;
  varcov = PowerExpand[FullSimplify[varcov /. facs]];
fact=-Last[MantissaExponent[Min[Abs[Select[Flatten[varcov], #!=0 &]]]]];
varcov=10^fact*varcov;

  head = Table[StyleForm[bt[[First[Flatten[Position[bt,svr1[[i]]]]],3]],
               FontSlant -> "Italic",  FontWeight -> "Bold"], {i,df}];
  ccout = Table[PowerExpand[varcov[[k,l]]/
                Sqrt[varcov[[k,k]]*varcov[[l,l]]]], {k,df},{l,df}];
If[ nuls==0, 
Print[TableForm[ccout, TableSpacing->{1,0.5}, TableAlignments->Center,
                       TableHeadings->{head,head}]];
      MatrixCharacteristics[ccout,"output correlation matrix"] ];

If[ okey=="Mix"||okey=="All", If[ nuls !=0, okey=Null; Goto[ret] ];
mixcm = Table[If[i>j , ccout[[i,j]], varcov[[i,j]]], {i,df}, {j,df}];

mix = Table[If[i>j, StyleForm[ccout[[i,j]], FontSlant -> "Italic"],
        StyleForm[varcov[[i,j]], FontWeight -> "Bold"]], {i,df}, {j,df}];

Print[Power[StyleForm["10", FontSize->9, FontSlant->"Italic",
                            FontWeight->"Bold", FontFamily->"Helvetica"],
            ToString[fact]], 
StyleForm[" times the values  of the relative  covariances  and variances 
(in boldface above and on the main diagonal); and correlation  coefficients
 (in italics below the main diagonal) :",
FontSize->9,FontSlant->"Italic",FontWeight->"Bold",FontFamily->"Helvetica"]];

Print[TableForm[mix, TableSpacing->{1,0.1}, TableAlignments->Center,
                       TableHeadings->{head,head}]] ];

        varcov = {varcov,10.^-fact};
        mixcm  = {mixcm,10.^-fact};
        If[ nuls!=0 || dm==1, Goto[ret] ];
        If[ flag==0, (cco=ccout; vrcv=varcov; mixo=mixcm),
                     (varcov={vrcv,varcov};
                      mixcm={mixo,mixcm};
                      ccout={cco,ccout}; Goto[ret]) ];
        If[ Min[Last[mch]]<0, flag = 1; ckey = 0; Goto[flg] ];
Label[ret];
Clear[mix];
If[ okey===Null, result = ccout,
    If[ okey=="All",
        If[ flag==0, result={ccout,First[varcov],First[mixcm],Last[mixcm]},
                     result=Table[{ccout[[i]],First[varcov[[i]]],
                                  First[mixcm[[i]]],Last[mixcm[[i]]]},{i,2}]],
        If[ okey=="CVM", result = varcov,
            If[ okey=="Mix", result = mixcm,
                             result = ccout;
Print[TableForm[{"Covariator: OUTPUT-keys: \"All\", \"CVM\", \"Mix\""}]] ]]]];

Label[end];   Return[{head,result}] ]


 MatrixCharacteristics[matr___List, headtxt___String] :=

            Module[ {cc=matr, ht=headtxt, cn=Null,det=Null,minr=Null,
                     eigvcc={},mineig=Null,cmplx},

  If[ !MatrixQ[cc], 
      Print["MatrixCharacteristics:  Input list is not a matrix"]; Goto[ret] ];
  If[ Dimensions[cc][[1]]!=Dimensions[cc][[2]],
      Print["MatrixCharacteristics:  Input matrix is not a square matrix"];
      Goto[ret] ];
  If[ ht === Null, ht = "the matrix" ];
  Print[StyleForm[" Characteristics of " <> ht, FontSize->10,
                    FontSlant->"Italic", FontWeight->"Bold",
                    FontFamily->"Helvetica"]];
  eigvcc = Eigenvalues[cc];
  If[ MatrixQ[cc,NumberQ],
      mineig = Min[Abs[eigvcc]];
      cn = If[Min[mineig] == 0, \[Infinity], Max[Abs[eigvcc]]/mineig],
      cn = "(Not evaluated - matrix is not numeric)" ];

  Print[StyleForm[" 1) Condition number = ",FontSize->10,FontSlant->"Italic",
                       FontWeight->"Bold",  FontFamily->"Helvetica"],
        StyleForm[cn,FontSize->10, FontSlant->"Italic", FontWeight->"Bold",
                                   FontFamily->"Helvetica"]];

  det = Times @@ eigvcc;
  Print[StyleForm[" 2) Determinant = ", FontSize->10, FontSlant->"Italic",
                    FontWeight->"Bold", FontFamily->"Helvetica"],
        StyleForm[det, FontSize->10, FontSlant->"Italic", FontWeight->"Bold",
                       FontFamily->"Helvetica"]];

   If[ !MatrixQ[cc,NumberQ], Goto[ret] ];
   cmplx = Cases[eigvcc, x_ /; Head[x] == Complex];
   If[ cmplx=!={}, Print["Vector of the matrix eigenvalues has complex components:"];
                   Print[cmplx]; Goto[ret] ];
   mineig = Min[eigvcc];
   If[ mineig<0, 
       Print[StyleForm[" 3) WARNING:  the matrix is non-positive semidefinite",
                       FontSize->10, FontSlant->"Italic", FontWeight->"Bold", 
                       FontFamily->"Helvetica"]];
       Print["   (Minimum eigenvalue of this matrix is ",mineig,")"],

       minr=If[ mineig == 0, \[Infinity], 
                Ceiling[Log[10,(First[Dimensions[eigvcc]]-1)/(2*mineig) ]] ];
             Print[StyleForm[" 3) Rounding criterion = ", FontSize->10,
             FontSlant->"Italic", FontWeight->"Bold",FontFamily->"Helvetica"],
                   StyleForm[minr, FontSize->10, FontSlant->"Italic",
                             FontWeight->"Bold", FontFamily->"Helvetica"]] ];

Label[ret]; 
Return[{cn,det,minr,eigvcc}] ]


 UM[exp_] := Module[ {goal=Null}, UM[exp,goal] ];

(* under study *)
 UM[exp_,goal___] :=
             Module[{expr=exp,gol=goal,subs={},
                     unisub={Newton -> kGram*Meter/Second^2,
                             Joule -> kGram*Meter^2/Second^2,
                             Watt -> kGram*Meter^2/Second^3,
                             Coulomb -> Ampere*Second,
                             Volt -> kGram*Meter^2/(Ampere*Second^3),
                             Tesla -> kGram/(Ampere*Second^2),
                             Weber -> kGram/(Ampere*Meter^2*Second^2)}},
 If[ gol=!=Null, If[ gol==1, Goto[end] ], unisub={}; Goto[end] ];
Print["UnitsManipulator:
 Sorry, computational substitutions are in elaboration now"];
unisub = {};
Label[end];
     If[ MemberQ[expr,eV,Infinity]||
                     MemberQ[expr,MeV,Infinity]||MemberQ[expr,GeV,Infinity],
     If[ MemberQ[expr,kGram,Infinity]||MemberQ[expr,kGram,Infinity], Print[
         "UnitManipulator::  Expression contains \"eV\" and \"kGram\""] ];
     If[ MemberQ[expr,eV,Infinity]&&
                     MemberQ[expr,MeV,Infinity]&&MemberQ[expr,GeV,Infinity],
         expr=expr/.eV->10^-6 MeV/.GeV->10^3 MeV,
         If[ MemberQ[expr,eV,Infinity]&&MemberQ[expr,MeV,Infinity],
                                                expr=expr/.eV ->10^-6 MeV ];
         If[ MemberQ[expr,eV,Infinity]&&MemberQ[expr,GeV,Infinity],
                                                expr=expr/.eV ->10^-9 GeV ];
         If[ MemberQ[expr,GeV,Infinity]&&MemberQ[expr,MeV,Infinity],
                                                expr=expr/.MeV->10^-3 GeV ]]];
 If[ MemberQ[expr,Hertz,Infinity]||MemberQ[expr,kHertz,Infinity],
     If[ MemberQ[expr,Hertz,Infinity]&&MemberQ[expr,kHertz,Infinity],
         expr=expr/.Hertz->10^-3 kHertz; If[ MemberQ[expr,Second,Infinity],
                                             expr=expr/.kHertz->10^3 Second^-1 ],
         If[ MemberQ[expr,Second,Infinity], expr=expr/.kHertz->10^3 Second^-1/.
             Hertz->Second^-1 ] ] ];
 expr = expr/.unisub;
 Return[expr]  ]


(* under study *)
 SetPower[number_,power_] := 
          Module[{n=number, m=power, in, im, n1, ms, str},
 
               mp = MantissaExponent[n];
               n1 = First[mp] * 10^(Last[mp]-m);
               ms = SuperscriptBox[10,m]//DisplayForm;
              str = If[ m==0, ToString[AccountingForm[n1]],
                              ToString[AccountingForm[n1]]<>""<>"\[Times]"ms];
 Return[str] ]


 Access[baslis:{___List},name_String]:=

   Module[ {bcl=baslis, nast=name, dim, val, gbt=Global`basicTable,
            slv=299792458,magc},

           dim = First[Dimensions[bcl]];
           val = Table["Exact",{i,3}];

           magc = 4 Pi 10^-7 Volt Second / Ampere / Meter;
           If[ Position[bcl,"speedOfLightInVacuum"] != {},
               slv = ToExpression[StringReplace[bcl[[First[Flatten[Position[bcl,
                                       "speedOfLightInVacuum"]]],4]]," "->""]],
               If[ Head[gbt] === List,
               slv = ToExpression[StringReplace[gbt[[First[Flatten[Position[gbt,
                                       "speedOfLightInVacuum"]]],4]]," "->""]] ] ];
           Do[ If[ nast==bcl[[i,2]], 
(* 1 *)           (If[ nast=="accelerationOfGravity", 
                       val[[1]]=ToExpression[StringReplace[bcl[[i,4]]," "->""]]*
                                Meter Second^-2;
                       Goto[ret] ];
(* 2 *)            If[ nast=="electricConstant",
                       val[[1]]=1/(magc*(slv Meter/Second)^2) /.
                                  Ampere -> Coulomb Second^-1;
                       Goto[ret] ];
(* 3 *)            If[ nast=="impedanceOfVacuum",
                       val[[1]]=magc*slv Meter/Second;
                       Goto[ret] ];
(* 4 *)            If[ nast=="magneticConstant", 
                       val[[1]] = magc;
                       Goto[ret] ];
(* 5 *)            If[ nast=="speedOfLightInVacuum", 
                       val[[1]]=ToExpression[StringReplace[bcl[[i,4]]," "->""]]*
                                Meter Second^-1;
                       Goto[ret] ];

                       val=ValueSyntax[bcl[[i]]];  Goto[ret]),
                                     Continue[] ],  {i,1,dim}];

               Print[" Access: ",nast,"  is absent in the input table"];
               Return[];

Label[ret];    Return[val]   ]


 ValueSyntax[subl_List]:=

      Module[ {sl=subl,val,sl5,leng,dash,leps,rips,tims,s1,s2,pw,pw1,
                       pnts,pnt1=0,pnt2, str1,str2,nprec,j},
              val = Table[0,{i,3}];
              If[ Dimensions[sl[[4]]] != {}, sl[[4]]=First[sl[[4]]] ];
              sl5 = ToExpression[
                    StringReplace[sl[[5]],
                                  {"mol" -> "Mole",     "mbarn" -> "mbarn",
                                   "fm"  -> "fMeter",   "km"    -> "kMeter",
                                   "cm"  -> "cMeter",   "m"     -> "Meter",
                                   "s"   -> "Second",   "kg"    -> "kGram",
                                   "g"   -> "Gram",     "pc"    -> "Parsec",
                                   "N"   -> "Newton",   "J"     -> "Joule",
                                   "Wb"  -> "Weber",    "K"     -> "Kelvin",
                                   "A"   -> "Ampere",   "C"     -> "Coulomb",
                                   "eV"  -> "eV",       "V"     -> "Volt",
                                   "T"   -> "Tesla",    "W"     -> "Watt",
                                   "Hz"  -> "Hertz",    "Gyr"   -> "GYear"}]];

              If[ sl[[5]]==" "||sl[[5]]=="", sl5=1 ];
              sl[[4]] = StringReplace[sl[[4]], " " -> ""];
              leng = StringLength[sl[[4]]];
              If[ StringMatchQ[sl[[4]],"*(*)*"] == True, Goto[interval] ];

              nprec = leng;
              If[ StringMatchQ[sl[[4]],"*.*"] == True, nprec=nprec-1 ];
              If[ StringPosition[sl[[4]],"\[Times]"] != {},
                  nprec = StringPosition[sl[[4]],"\[Times]"][[1,1]] - 2];

              val[[1]] = SetPrecision[ToExpression[sl[[4]]], nprec] * sl5;
              Goto[ret];

Label[interval];   If[ StringTake[sl[[4]],{1}] != "(", Goto[inexact] ];
                       dash = StringPosition[sl[[4]],"-"][[1,1]];
                       rips = StringPosition[sl[[4]],")"][[1,1]];
                       tims = 
                   If[ First[Dimensions[StringPosition[sl[[4]],"\[Times]"]]]==0,
                       0,  StringPosition[sl[[4]],"\[Times]"][[1,1]]  ];
                   str1 = StringTake[sl[[4]],{2,dash-1}];
                   leng1 = StringLength[str1];
                   str2 = StringTake[sl[[4]],{dash+1,rips-1}];
                   leng2 = StringLength[str2];
                   If[ StringMatchQ[sl[[4]],"*.*"] == False, 
                   nprec = Dimensions[IntegerDigits[IntegerPart[
                           (ToExpression[str1]+ToExpression[str2])/2 ]]][[1]],
                   nprec = Max[leng1,leng2]-1;  If[StringTake[str1,{1}]=="0",
                   nprec = nprec - 1; If[leng1>leng2, nprec = nprec - 1] ] ];

If[ StringTake[str1,{1}]=="0" && leng1 > leng2,
    If[ OddQ[ToExpression[StringTake[str1,{leng1}]]]==True,
             nprec = nprec+1 ],
                   If[ (EvenQ[ToExpression[StringTake[str1,{leng1}]]]==True &&
                         OddQ[ToExpression[StringTake[str2,{leng2}]]]==True)||
                       ( OddQ[ToExpression[StringTake[str1,{leng1}]]]==True &&
                        EvenQ[ToExpression[StringTake[str2,{leng2}]]]==True),
                        nprec = nprec+1 ]  ];
                   s1 = SetPrecision[ToExpression[str1], nprec];
                   s2 = SetPrecision[ToExpression[str2], nprec];
                   pw = If[ tims==0, 1,
                            ToExpression[StringTake[sl[[4]],{tims+1,leng}]] ];
             val[[1]] = (s1+s2)/2 * pw * sl5;
             val[[2]] = (s2-s1)/2 * pw * sl5;
             val[[3]] = SetPrecision[val[[2]]/val[[1]], 2];
             Goto[ret];

Label[inexact];        leps = StringPosition[sl[[4]],"("][[1,1]];
                       rips = StringPosition[sl[[4]],")"][[1,1]];
                       pnts = StringPosition[sl[[4]],"."];
                       pnt2 = First[Dimensions[pnts]];
                       If[ pnt2==2 || pnt2==0, pnts=0, pnt1=pnts[[1,1]];
                                                       pnts=leps-pnt1-1 ];
                       tims = 0;
                       If[ StringPosition[sl[[4]],"\[Times]"] != {},
                       tims = StringPosition[sl[[4]],"\[Times]"][[1,1]] ];
                   j = 0;
                   nprec = leps-1;
                   If[ StringTake[sl[[4]],{1}]=="-", nprec-=1; j+=1];
                   If[ pnt1 != 0 || pnt2==2, nprec-=1 ];
Label[precis];    
                   j += 1;
                   If[ StringTake[sl[[4]],{j}]=="0", nprec-=1; Goto[precis] ];
                   If[ StringTake[sl[[4]],{j}]==".", Goto[precis] ];
                   s1 = SetPrecision[
                        ToExpression[StringTake[sl[[4]],{1,leps-1}]], nprec];

                   nprec = rips-leps-1;
                   str2 = StringTake[sl[[4]],{leps+1,rips-1}];
                   If[ StringPosition[str2,"."] != {}, nprec=nprec-1 ];
                   s2 = SetPrecision[ToExpression[str2], nprec];
      If[ tims==0, pw = 1,
                   pw = ToExpression[StringTake[sl[[4]],{tims+1,leng}]] ];
                   pw1= 10.^-pnts;
             val[[1]] = s1 * pw * sl5;
             val[[2]] = SetPrecision[s2*pw1*pw, nprec] * sl5;
             val[[3]] = If[ val[[1]]=!=0.,SetPrecision[Abs[val[[2]]/val[[1]]],2],
                                     "",SetPrecision[Abs[val[[2]]/val[[1]]],2] ];
Label[ret];  
                Return[val]     ]


 MakeCM[database___List] :=
      Module[ { print=Null },
              MakeCM[database,print]  ]

 MakeCM[database_List,print_] :=
      Module[ { db=database,cct,prt=print,pos, gdb=Global`dataBase,
                cm={},n, n2, dim, dim2, i, j, k },

            If[ Depth[db]==2, db=NewListFPC[db, gdb] ];

                If[ First[Dimensions[db]] != 3, Goto[er2] ];
                pos = Flatten[Position[Map[
                      Access[db[[1]],#][[2]]&, db[[1,All,2]]],"Exact"]];
                If[ pos =!= {}, db = DeleteFPC[pos, db] ];
                cct = Last[db];
                n = First[Dimensions[cct]];
                If[ n<=0, Goto[err] ];
                If[ cct[[1,3]]==cct[[1,4]], dim=1; Goto[cnt] ];
                n2 = 2*n;
                dim = 0;
Label[nxt];     dim=dim+1; dim2 = dim(dim-1);
             If[dim2 > n2, Goto[er1]];
             If[dim2 == n2, Goto[cnt], Goto[nxt]];
Label[cnt];     cm = Map[ToString,SetPrecision[IdentityMatrix[dim],4],{2}];
              head = Table[0,{i,dim}];
              If[ dim==1, head[[1]]=cct[[1,3]]; cm[[1,1]]=cct[[1,5]]; 
                          cm=Append[cm,{}]; Goto[ret] ];
                 k = 0;
             Do[ Do[ k+=1; cm[[i,j]] = cct[[k,5]];
                           cm[[j,i]] = cct[[k,5]],
                                        {j,i+1,dim} ];
                              head[[i]] = cct[[k,3]],
                                        {i,1,dim-1} ];
                              head[[dim]] = cct[[n,4]];
Label[ret];
             If[prt=="print", Print[TableForm[cm, TableSpacing->{1,0.2},
                            TableAlignments->Right,
                            TableHeadings->{head,head}]],
                            If[ prt=!=Null,
                                Print[" 2-d parameter must be 'print'"]] ];
             If[ MatrixQ[cm], Null, cm=Drop[cm, -1] ];
             Return[ToExpression[cm]];

Label[er2];  Print["  MakeCM:  Argument is not FPC database"];
             Return[];

Label[er1];  Print["  MakeCM:  You are using non-full CC-table"];
             Return[];

Label[err];  Print["  MakeCM:  CC-table dimension is <= 0"];
             Return[]      ]


 PairCC[ncfpc1_,ncfpc2_,db:{___List}] :=
      Module[ {nc1=ncfpc1,nc2=ncfpc2, lst=First[db], cct=Last[db],
               dim,dct,c1,cc1,c2,cc2, errstr,x,i,j,dm},

                dim = First[Dimensions[lst]];
                dct = First[Dimensions[cct]];
                errstr = "No such constant: ";

           If[ NumberQ[nc1] == True,
               If[ nc1<=dim, c1=nc1; Goto[nxt], Goto[err1] ],
               Do[ If[ lst[[i,2]]===nc1, c1=i;  Goto[nxt] ],
                                                {i,1,dim} ];
                   Goto[err1] ];
Label[nxt];
           If[ NumberQ[nc2] == True,
               If[ nc2<=dim, c2=nc2; Goto[do], Goto[err2] ],
               Do[ If[ lst[[i,2]]===nc2, c2=i; Goto[do] ],
                                               {i,1,dim} ];
                   Goto[err2] ];
Label[do];
               xxxp = {"","","","","1.000"};
           If[ c1==c2, xxxp[[1]]=c1;         xxxp[[2]]=c2;
                       xxxp[[3]]=lst[[c1,3]];xxxp[[4]]=xxxp[[3]];
                       Goto[end] ];

                      cc1 = c1;  cc2 = c2;
           If[ c1>c2, cc1 = c2;  cc2 = c1 ];
             dm = dim - 1;
             i = 1;
Label[lp1];
               If[ cct[[i,1]] != cc1, i += dm;  dm -= 1; Goto[lp1],
                   j = i-1;
Label[lp2];  j+=1;
               If[ cct[[j,2]] != cc2, Goto[lp2],
                                      xxxp = cct[[j]]   ] ];

           If[ c1>c2,x=xxxp[[1]];xxxp[[1]]=xxxp[[2]];xxxp[[2]]=x;
                     x=xxxp[[3]];xxxp[[3]]=xxxp[[4]];xxxp[[4]]=x;
                     Goto[end], Goto[end] ];
Label[end];
               Return[xxxp];

Label[err1];   Print[errstr,nc1];
               Return[{nc1,Null,Null,Null,Null}];
Label[err2];   Print[errstr,nc2];
               Return[{nc1,nc2,Null,Null,Null}]  ];


 GroupCC[ncfirst_,ncgr_List,db:{___List}] :=
      Module[ { print=Null },
              GroupCC[ncfirst,ncgr,db,print]  ]

 GroupCC[ncfirst_,ncgr_List,db:{___List},print_] :=
      Module[ { nc1=ncfirst, ncg=ncgr, lst=First[db], cct=Last[db],
                prt=print, nc2, dgr, dim, d, i, j },

                d   = First[Dimensions[lst]];
           If[ NumberQ[nc1] == True,
               If[ nc1>d, Goto[err1]  ],
               Do[ If[ lst[[i,2]]===nc1, nc1=i; Goto[go1] ],
                                                  {i,1,d} ];
                   Goto[err1] ];
Label[go1];
                dgr = First[Dimensions[ncg]];
                dim = Last[Dimensions[cct]];
                nct=Table[0,{i,1,dgr},{j,1,dim}];
            Do[ nc2=ncg[[i]]; nct[[i]]=PairCC[nc1,nc2,db];
                If[Last[nct[[i]]]==Null, nct[[i,3]]="<-No";
                                         nct[[i,4]]="such";
                                         nct[[i,5]]="constant"],
                                                       {i,dgr}];

           chead = {"i","j",SubscriptBox["x","i"]//DisplayForm,
                            SubscriptBox["x","j"]//DisplayForm,
                            SubscriptBox["r","ij"]//DisplayForm};
 If[ prt=="print",Print[TableForm[nct,TableAlignments -> {Right},
                                      TableSpacing -> {1,2},
                                      TableHeadings -> {None,chead}]] ];
 Return[nct];

Label[err1]; Print[" GroupCC: The first constant is absent in database"];
             Return[Null]  ]


InsertFPC[lis_List,db:{___List}] :=
          Module[ {ccdef=" ", newdb},
                  newdb=InsertFPC[lis,db,ccdef];
                  Return[newdb]  ]

InsertFPC[lis_List,db:{___List},ccdef_] :=
          Module[ {ndb},
                  ndb = InsertDB[lis,db[[1]],db[[2]],db[[3]],ccdef];
                  Return[ndb] ]

InsertDB[lisin_List,lst:{___List},stn_List,cct:{___List},ccd_] :=
      Module[ {lis=lisin,bdim,dim,i,defnam5=ccd,n0=1,n1,n2,n3,nn,i1,i2,
               list,xx,yy,j,
               addns,cc0=cct,k,date, ins, namlst, bl={},sn={},ccn},

              If[ First[Dimensions[lis]]==5, addns=lis[[2]], 
                  If[ TrueQ[lis[[6]]==Null], addns=lis[[2]], 
                      addns=lis[[6]] ]; lis=Drop[lis,-1] ];
              If[ cct[[1,3]] != cct[[1,4]], Goto[big] ];
              If[ lst[[1,1]]===Null, bl=Append[bl,lis]; bl[[1,1]]=1; 
                                     sn=Append[sn,addns]; 
                            cc0[[1,5]]="1.000"; cc0[[1,1]]=cc0[[1,2]]=1;
                            cc0[[1,3]]=cc0[[1,4]]=lis[[3]]; ccn=cc0;
                            Goto[cend1] ];
Label[big];
              ins=lis[[2]];
              namlst=lst[[All,2]];
              If[ MemberQ[namlst,ins], Print["InsertFPC::  ",ins,
                  "exist already"]; bl=lst; sn=stn; ccn=cct;
                  n0=Length[cc0]; Goto[cend1] ];

              list = Table[0,{i,5}];
              j = 0;
              If[ TrueQ[lis[[1]]==Null],j=StringAO[ins,namlst],j=lis[[1]] ];

              list[[1]] = j;  Do[ list[[i]]=lis[[i]], {i,2,5} ];

                bdim = First[Dimensions[lst]];
                dim  = bdim + 1;
                bl = Insert[lst,list,j];
                If[ j==dim, Goto[end] ];
                Do[ bl[[i,1]]=bl[[i,1]]+1, {i,j+1,dim} ];
Label[end];
                sn = Insert[stn, addns,j];

       If[ cc0[[1,3]]==cc0[[1,4]], ccn=Table[Null,{i,1},{j,5}];
             ccn[[1,1]]=bl[[1,1]]; ccn[[1,2]]=bl[[2,1]];
             ccn[[1,3]]=bl[[1,3]]; ccn[[1,4]]=bl[[2,3]];
             ccn[[1,5]]=defnam5;   Goto[cend1] ];

                n0 = bdim(bdim-1)/2;
            If[ j==bdim+1,Goto[c2]];
                n2 = bdim-j+1;
                n1 = n2*(bdim-j)/2;
                i1 = n0-n1+1;
                xx = Take[cc0,{i1,n0}]/. (e_/; NumberQ[e]) -> e+1;
              If[ j != 1, Goto[c0] ];
                  yy = Take[cc0,{1,bdim-1}];
                  yy = Prepend[yy,yy[[1]]];
                  yy[[1,4]] = yy[[1,3]];
                  yy[[1,3]] = list[[3]];
                  yy[[1,5]] = defnam5;
              Do[ yy[[i,2]] = yy[[i,2]] + 1;
                  yy[[i,3]] = list[[3]];
                  yy[[i,5]] = defnam5,
                                                   {i,2,n2}];
                Goto[cend];

Label[c0];      i2 = i1-n2;
           yy = Take[cc0,{i2,i1-1}];
              Do[ yy[[i,1]] = j; yy[[i,2]] = yy[[i,2]] + 1;
                  yy[[i,3]] = list[[3]];
                  yy[[i,5]] = defnam5,
                                                   {i,1,n2}];
                 n3 = (j-1)(2*bdim-j)/2;
                 nn = n3 + (j-1);
           ccn = Take[cc0,{1,n3}];

            i = 0;
Label[c1];  i = i+1;  If[ i > nn, Goto[cend] ];
                      If[ ccn[[i,2]]<j, Goto[c1] ];
                      If[ ccn[[i,2]]>j, ccn[[i,2]]=ccn[[i,2]]+1,
    ccn = Insert[ccn,{ccn[[i,1]],j,ccn[[i,3]],list[[3]],defnam5},i];
                           i=i+1;  ccn[[i,2]]=ccn[[i,2]]+1    ];
                      Goto[c1];
Label[c2];
           ccn = Append[cc0,{bdim,j,cc0[[n0,4]],list[[3]],defnam5}];

                       i = 0;
Label[c3];       i = i+1; k = n0-(i-2)(i+1)/2;
                 If[ k-1<=0, Goto[cend1] ];
                 ccn = Insert[ccn,ccn[[k-1]],k];
                 ccn[[k,2]] = j;
                 ccn[[k,4]] = list[[3]];
                 ccn[[k,5]] = defnam5;
             Goto[c3];
Label[cend];     If[ j==1, ccn=Join[yy,xx], ccn = Join[ccn,yy,xx] ];
                     xx=Drop[xx,n1];  yy=Drop[yy,n2];

Label[cend1];    Drop[cc0,n0];
  
     Return[{bl,sn,ccn}]  ];


StringAO[stg_, aol_List] :=
      Module[ {cl=aol,dlst, numb=0, icds,ilth, ncds,nlth},

                dlst = First[Dimensions[cl]];
                icds = ToCharacterCode[ToLowerCase[stg]];
                ilth = StringLength[stg];
                
       Do[ numb=i; ncds=ToCharacterCode[ToLowerCase[cl[[i]]]];
           nlth=StringLength[cl[[i]]];      dj=Min[ilth,nlth];

           Do[ If[ icds[[j]]==ncds[[j]], Continue[] ];
               If[ icds[[j]]< ncds[[j]], Goto[exit], Goto[nextc] ],
               {j,dj} ];

           If[ ilth< nlth, Goto[exit] ]; Label[nextc],
           {i,dlst} ];

       numb = dlst + 1;
Label[exit];
             Return[numb]  ]


 DeleteFPC[namecode_,db:{___List}] :=
             DeleteDB[namecode,db[[1]],db[[2]],db[[3]]];

 DeleteDB[namecode_,lst:{___List},stn_List,cct:{___List}] :=

      Module[ { lnc,nc,cnc=1,xc, bl=lst,sn=stn,ccn=cct, k,miss={},
                j,dim,d1,dc0,dc, n1,ne, prt, dnc=0 },

       If[ ListQ[namecode], lnc=namecode, lnc={namecode} ];

       If[ sn==={}, bl={{Null,Null,Null,Null,Null}};
                    ccn={{Null,Null,Null,Null,Null}}; miss=lnc;
                    Goto[ret] ];

       dim = First[Dimensions[bl]];
       dnc = First[Dimensions[lnc]];
       k = 0;
Label[nxt];  k+=1;  If[ k>dnc,
                        If[ lnc=!={}, lnc=Reverse[Sort[lnc]]; Goto[clis],
                                      Goto[ret] ] ];
             xc=lnc[[k]];
             If[ NumberQ[xc]==True, 
                 If[ xc>dim, Goto[warn], Goto[nxt]],
                             j = Flatten[Position[bl,xc]];
                             If[ j==={}, Goto[warn], lnc[[k]]=First[j];
                                                     Goto[nxt] ] ];
Label[warn]; miss = Append[miss,lnc[[k]]];
             lnc  = Delete[lnc,k];
             k-=1; dnc-=1; Goto[nxt];

Label[clis]; j = lnc[[cnc]];

Label[onec];
       dim = First[Dimensions[bl]];
       If[ dim==1, bl={{Null,Null,Null,Null,Null}}; sn={};
                   ccn={{Null,Null,Null,Null,Null}}; Goto[ret] ];
       d1 = dim-1;
       dc0= dim*d1/2;
       dc = dc0-d1;

       sn = Delete[sn, j];
       bl = Delete[bl, j];
       If[ j<dim, prt = Table[{i,1}, {i,j,d1}];
                   bl = MapAt[#-1 &, bl, prt] ];

       If[ j==1, ccn = Drop[ccn,d1];
                 ccn=ccn /.(e_/;NumberQ[e])->e-1; Goto[end] ];

       n1 = (j-1)(2*dim-j)/2 + 1;
       ne = j*(2*dim-j-1)/2;
       ccn = Drop[ccn, {n1,ne}];
       prt = Table[j-1, {i, j-1}];
       Do[ prt[[i]] = prt[[i-1]]+dim-i, {i, 2, j-1} ];
       prt = Partition[prt, 1];
       ccn = Delete[ccn, prt];
                 ccn=ccn /.(e_/;NumberQ[e] && e>j)->e-1;
Label[end];
       If[ cnc < dnc, cnc = cnc+1; Goto[clis] ];

Label[ret];
   If[miss=!={}, Print["DeleteFPC:: No constant(s) in current DB: ",miss]];
   If[ First[Dimensions[sn]]==1, ccn={{1,1,bl[[1,3]],bl[[1,3]],"1.000"}} ];

    Return[{bl,sn,ccn}]  ];


 Copy[clis_, dbs:{___List}, dbrcvr:{___List}] :=

      Module[ { cl, nc, ds, dbr=dbrcvr, pos, k, ins, pc, npc, c1, c2 },

         If[ ListQ[clis], cl=clis, cl={clis} ];
         If[ cl==={} || cl==={Null}, Goto[end] ];
             nc = Length[cl];
             ds = Length[dbs[[1]]];
         Do[ k=cl[[i]]; If[ NumberQ[k], If[ k>ds, Print[
             "Copy:: Constant code ",k," is absent in DB-source"];
                            Continue[], k=dbs[[1,k,2]] ] ];
             pos=Position[dbs[[1]],k];  If[ pos==={}, Print[
             "Copy:: Constant ",k," is absent in DB-source"];
                                                      Continue[] ];
             If[ Position[dbr[[1]],k]=!={}, 
                 Print["Copy::  ",k," exists in DB-recipient already"];
                 Continue[] ];
             pos=First[Flatten[pos]]; ins=dbs[[1,pos]];
             ins[[1]]=Null; ins=Append[ins,dbs[[2,pos]]];
             dbr=InsertFPC[ins,dbr];

             If[ dbr[[3,1,3]]==dbr[[3,1,4]], Continue[] ];
             pc = Map[First[#]&,Position[dbr[[3]]," "]];
             If[ pc==={}, Continue[] ];
             npc = Length[pc];
             Do[k=pc[[j]]; c1=dbr[[3,k,1]]; c1=dbr[[1,c1,2]];
                           c2=dbr[[3,k,2]]; c2=dbr[[1,c2,2]];

                If[MemberQ[dbs[[1]],c1,2] && MemberQ[dbs[[1]],c2,2],
                   dbr[[3,k,5]]=Last[PairCC[c1,c2,dbs]],
                   If[ Last[Access[dbr[[1]],c1]]=="Exact"||
                       Last[Access[dbr[[1]],c2]]=="Exact",
                       dbr[[3,k,5]]="0.000", dbr[[3,k,5]]="Null" ] ], 
                {j,npc} ],             
                            {i,nc} ];

Label[end];   Return[dbr]  ]


 Shift[codname_,newpos_,database:{___List}] :=
      Module[ {nc=codname,np=newpos,cl=database[[1]],nl=database[[2]],
               cct=database[[3]], dim,dct,j,xm,j0,jm,k,i,min,max,nmax},

                dim = First[Dimensions[cl]];
                If[ np>dim, Print["  Wrong new position! (code: ",np,")"];
                            Return[{}] ];

                If[NumberQ[nc] == True, 
                   If[ nc<=dim, j=nc, Goto[err] ],

                       j = Flatten[Position[cl,nc]];
                       If[j==={}, Goto[err], j=First[j]] ];
                If[ j==np, Goto[ret] ];
                If[ j>np, min=np; max=j; k=1, min=j; max=np; k=-1 ];
                xm = cl[[j]];
                xm[[1]] = np;
                cl = Delete[cl,j];
                cl = Insert[cl,xm,np];
                j0 = j; jm = np-1;
                If[ j>np, j0=np+1; jm=j ];
                Do[ cl[[i,1]] += k, {i,j0,jm} ];

                xm = nl[[j]];
                nl = Delete[nl,j];
                nl = Insert[nl,xm,np];

                nmax = max(2 dim-max-1)/2;
                i = 0;
Label[nxt];
                i += 1;  If[ i>nmax, Goto[sort] ];
                If[ cct[[i,1]]<min,
                    If[ cct[[i,2]]<min, Goto[nxt],
                        If[ cct[[i,2]]>max, Goto[nxt],
                            If[ cct[[i,2]]==j, cct[[i,2]]=np; Goto[nxt] ];
                            cct[[i,2]] += k; Goto[nxt] ] ] ];

                If[ cct[[i,1]]!=j, cct[[i,1]]+=k;
                    If[ cct[[i,2]]>max, Goto[nxt],
                        If[ cct[[i,2]]==j, cct[[i,2]]=np, cct[[i,2]]+=k ];
                        If[ cct[[i,1]]>cct[[i,2]], Goto[perm], Goto[nxt] ]],
                        cct[[i,1]] = np;
                        If[ j==max, Goto[nxt],
                                    If[ cct[[i,2]]>max, Goto[nxt],
                                        cct[[i,2]]+=k; Goto[perm] ] ] ];
Label[perm];
                xm = cct[[i,1]];  cct[[i,1]] = cct[[i,2]];  cct[[i,2]] = xm;
                xm = cct[[i,3]];  cct[[i,3]] = cct[[i,4]];  cct[[i,4]] = xm;
                Goto[nxt];
Label[sort];
                cct = Sort[cct];
Label[ret];
                Return[{cl,nl,cct}];
Label[err];
         Print[" No such constant! (code or name is wrong: ",nc,")"];
         Return[{}]    ]


 NewListFPC[nclist___List,db:{___List}] :=
      Module[ { ncl=nclist,lst,sn,cct, newl={{Null,Null,Null,Null,Null}},
                bcod={},ncct={{Null,Null,Null,Null,Null}}, pos,
                d,n,dim,dim2,nc, i,j,k,m, d1,d2,ifirst,ilast },

                If[ ncl==={}, Goto[end] ];
                lst  = First[db];
                sn   = db[[2]];
                cct  = Last[db];
                d    = Length[ncl];
                d1   = d-1;
                n    = d d1/2;
                Do[If[StringQ[ncl[[i]]], nc=ncl[[i]]; pos=Position[lst,nc];
                                         If[ pos==={}, Goto[err] ];
                                         ncl[[i]]=First[Flatten[pos]]],{i,d}];
                ncl  = Sort[ncl];
                dim  = First[Dimensions[lst]];
                d2   = 2 dim;
                dim2 = Last[Dimensions[lst]];

                newl = Table["",{i,1,d},{j,1,dim2}];
       If[ n>0, ncct = Table["",{i,1,n},{j,1,5}] ];
                bcod = Table[ 0,{i,1,d} ];

                Do[ nc = ncl[[i]]; If[ nc>dim || nc<1, Goto[err] ];
                    newl[[i]] = lst[[nc]]; bcod[[i]] = lst[[nc,1]];
                    newl[[i,1]] = i,                     {i,1,d} ];

              If[ d==1, ncct[[1,1]]=ncct[[1,2]]=1;
                        ncct[[1,3]]=ncct[[1,4]]=newl[[1,3]];
                        ncct[[1,5]]="1.000"; Goto[end] ];
              i = 0;  k = 0;
Label[nxt1];  i = i+1; If[ i > d1, Goto[end]];  j = i;
Label[nxt2];  j = j+1; If[ j > d,  Goto[nxt1],  k = k+1;
                                  c1 = bcod[[i]]; c2 = bcod[[j]];
                                  If[ c1==c2, Goto[nxt2] ];
                                  If[ c1>c2, c1=c2; c2=bcod[[i]] ];
                            ncct[[k,1]]=i; ncct[[k,3]]=newl[[i,3]];
                            ncct[[k,2]]=j; ncct[[k,4]]=newl[[j,3]];
                            ifirst = (c1-1)(d2-c1)/2 + 1;
                            ilast  = c1(d2-c1-1)/2;
                            Do[ If[ c2==cct[[m,2]],
                              ncct[[k,5]] = cct[[m,5]]; Goto[nxt2]],
                                                 {m,ifirst,ilast} ];
                            Goto[err1]  ];

Label[end];
 If[ sn=!={} && bcod=!={}, bcod=Part[sn,bcod] ];
 Return[{newl,bcod,ncct}];

Label[err];   Print[" NewListFPC: No such constant: ",nc];
              Return[];

Label[err1];  Print[" NewListFPC: No corr. coef. for these constants ( ",
                    c1,", ",c2," )"];
              Return[]              ]


 PrintDB[db_List] := Module[ {tab=db, dim},

               dim = Dimensions[tab];
               If[ Head[tab[[1]]]=!=List, PrintSN[tab],
                   If[ Last[dim]!=5, PrintBT[tab[[1]]];
                                     PrintSN[tab[[2]]];
                                     PrintCC[tab[[3]]],
                       If[ NumberQ[tab[[1,2]]]==False, PrintBT[tab],
                                                       PrintCC[tab] ]]];

Return[]  ]

 PrintBT[bt_List] := 
                    Module[ { head,xbt },

                    head = {"Code","Name","Symbol","Value","Units"};
 Print[StyleForm["Constant Table",
                             FontSize->11,
                             FontSlant->"Italic",
                             FontWeight->"Bold",
                             FontFamily->"Helvetica",
                             FontColor->RGBColor[0,0.5,0.999]]];
 xbt=Prepend[bt,head];
 Print[TableForm[xbt,  TableSpacing -> {1,0.5}]];

Return[]   ]

 PrintSN[sn_List] := 
                    Module[ { xsn=sn, dim, msn=Global`standardNames },

                    dim = First[Dimensions[xsn]];

                    If[ ValueQ[Global`standardNames]==True,
                        xsn = Table[If [NumberQ[xsn[[i]]]==True,
                                        msn[[xsn[[i]]]],
                                        xsn[[i]]], {i,dim}] ];

 Print[StyleForm["Constant Standard Names",
                             FontSize->11,
                             FontSlant->"Italic",
                             FontWeight->"Bold",
                             FontFamily->"Helvetica",
                             FontColor->RGBColor[0,0.5,0.999]]];
 Print[TableForm[xsn, TableDirections->Column,TableSpacing->{1,1},
                      TableAlignments->{Left},TableHeadings->Automatic]];

Return[]   ]

 PrintCC[ccn_List] :=
                    Module[ { chead },

                    chead = {"i","j",SubscriptBox["x","i"]//DisplayForm,
                            SubscriptBox["x","j"]//DisplayForm,
                            SubscriptBox["r","ij"]//DisplayForm};

 Print[StyleForm["Correlation Coefficient Table",
                             FontSize->11,
                             FontSlant->"Italic",
                             FontWeight->"Bold",
                             FontFamily->"Helvetica",
                             FontColor->RGBColor[0,0.5,0.999]]];

 Print[TableForm[ccn, TableAlignments -> {Right},
                      TableSpacing -> {1,2},
                      TableHeadings -> {None,chead}]];
Return[]   ]


 OutputDB[db:{___List}] :=
                    Module[ { bl=db[[1]],sn=db[[2]],ccn=db[[3]],date,
                              dim, msn=Global`standardNames },

                       date = Date[];
                       dim  = First[Dimensions[sn]];

OutputForm[" "]                                             >> Dbfpc_n.m;
OutputForm["(*  Data Base of the Fundamental Physical Constants"]
                                                            >>>Dbfpc_n.m;
OutputForm["    -----------------------------------------------  *)"]
                                                            >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["(* :Authors: V.V.Ezhela, V.N.Larin"]            >>>Dbfpc_n.m;
OutputForm["             --------------------- *)"]         >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["(*  Last file revision: "<>
                               ToString[date[[3]]]<>"."<>
                               ToString[date[[2]]]<>"."<>
                               ToString[date[[1]]]<>" *)"]  >>>Dbfpc_n.m;
OutputForm["(*  -----------------------------         *)"]  >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;          

OutputForm["(*   Last data revision:   *)"]                 >>>Dbfpc_n.m;
OutputForm["(*   -------------------   *)"]                 >>>Dbfpc_n.m;

 OutputForm[
 "(*   DATE:  "<>ToString[date[[3]]]<>"."<>
                 ToString[date[[2]]]<>"."<>
                 ToString[date[[1]]]<>" *)"]                >>>Dbfpc_n.m;
 
 OutputForm[
 "(*   TIME:  "<>ToString[date[[4]]]<>":"<>
                 ToString[date[[5]]]<>"  *)"]               >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["(*  --------------------                  *)"]  >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["(*   Number of the constants: "<>
                                  ToString[dim]]            >>>Dbfpc_n.m;
OutputForm["     Number of the correlation coefficients: "<>
                                  ToString[dim(dim-1)/2]]   >>>Dbfpc_n.m;
OutputForm["     --------------------------------------*)"] >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["date = Date[];"]                                >>>Dbfpc_n.m;
OutputForm["
 Print[\"     ***   DATE:  \",date[[3]],\".\",date[[2]],\".\",date[[1]]]"]
                                                            >>>Dbfpc_n.m;
OutputForm["
 Print[\"     ***   TIME:  \",date[[4]],\":\",date[[5]]]"]  >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm[" basicTable = "]                                >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
            bl                                              >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm[" (* PrintBT[basicTable]; *)"]                   >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm[" standardNames = "]                             >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
            sn = Table[If [NumberQ[sn[[i]]]==True,
                                   msn[[sn[[i]]]],
                                   sn[[i]]], {i,dim}];
            sn                                              >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm[" ccTable = "]                                   >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
            ccn                                             >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["    dataBase = {basicTable, standardNames, ccTable};"]
                                                            >>>Dbfpc_n.m;
OutputForm[" "]                                             >>>Dbfpc_n.m;
OutputForm["
(* =====================   END of Dbfpc.m   ====================== *)"]
                                                            >>>Dbfpc_n.m;
Return[]   ]

End[]

SetAttributes[ ErrorPropagator,      ReadProtected];
SetAttributes[ Covariator,           ReadProtected];
SetAttributes[ MatrixCharacteristics,ReadProtected];
SetAttributes[ UM,                   ReadProtected];
SetAttributes[ SetPower,             ReadProtected];
SetAttributes[ Access,               ReadProtected];
SetAttributes[ ValueSyntax,          ReadProtected];
SetAttributes[ MakeCM,               ReadProtected];
SetAttributes[ PairCC,               ReadProtected];
SetAttributes[ GroupCC,              ReadProtected];
SetAttributes[ InsertFPC,            ReadProtected];
SetAttributes[ InsertDB,             ReadProtected];
SetAttributes[ StringAO,             ReadProtected];
SetAttributes[ DeleteFPC,            ReadProtected];
SetAttributes[ DeleteDB,             ReadProtected];
SetAttributes[ Copy,                 ReadProtected];
SetAttributes[ Shift,                ReadProtected];
SetAttributes[ NewListFPC,           ReadProtected];
SetAttributes[ PrintDB,              ReadProtected];
SetAttributes[ PrintBT,              ReadProtected];
SetAttributes[ PrintSN,              ReadProtected];
SetAttributes[ PrintCC,              ReadProtected];
SetAttributes[ OutputDB,             ReadProtected];

Protect[ErrorPropagator,Covariator,MatrixCharacteristics,UM,SetPower,
        Access,ValueSyntax,MakeCM,PairCC,GroupCC, InsertFPC,InsertDB,
        StringAO,DeleteFPC,DeleteDB,Copy,Shift,NewListFPC,
        PrintDB,PrintBT,PrintSN,PrintCC,OutputDB, cC,
        mbarn,fMeter,kMeter,cMeter,Meter,Gram,kGram,Joule,eV,MeV,GeV,
        Watt,Ampere,Coulomb,Volt,Tesla,Weber,Hertz,kHertz,
        Mole,Kelvin,kParsec,GYear];

EndPackage[]